/******************************************************************************
 *
 * Header file for harness processes (process launcher) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/* Avoid to include the content of this header file twice. */
#ifndef HARNESS_PROCESSES_H
#define HARNESS_PROCESSES_H

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Starts a new process and executes a program.
 *
 * handle    = The process handle (return).
 * arguments = The program arguments (first argument is program name).
 * variables = The program variables (may be NULL).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_processes_execute (unsigned int *handle,
                               const char   *arguments,
                               const char   *variables);

/*
 * Starts a new process and executes a program.
 *
 * handle = The process handle (return).
 * argc   = The number of arguments.
 * envc   = The number of variables.
 * argv   = The array of arguments (argv[0] is program name).
 * envv   = The array of variables.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_processes_execute2 (unsigned int  *handle,
                                unsigned int   argc,
                                unsigned int   envc,
                                const char   **argv,
                                const char   **envv);

/*
 * Reads the standard output of a process into a buffer (non-/blocking).
 *
 * handle = The process handle.
 * length = The maximum buffer length to read.
 * buffer = The buffer to read into.
 * block  = The (non-)blocking flag (0 = non-blocking, 1 = blocking).
 * return = number of bytes read on success, -2 on closed pipe or -1 on any
 *          other error with errno set appropriately.
 */
int harness_processes_read (unsigned int  handle,
                            unsigned int  length,
                            void         *buffer,
                            unsigned int  block);

/*
 * Writes a buffer into the standard input of a process (blocking).
 *
 * handle = The process handle.
 * length = The length to write.
 * buffer = The buffer to write.
 * return = 0 on success, -2 on closed pipe or -1 on any
 *          other error with errno set appropriately.
 */
int harness_processes_write (unsigned int  handle,
                             unsigned int  length,
                             const void   *buffer);

/*
 * Sends a signal to a process.
 *
 * handle = The process handle.
 * signum = The signal to send.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_processes_signal (unsigned int handle,
                              int          signum);

/*
 * Waits for the exit of a process, but does not
 * remove the entry in the kernel process table.
 *
 * handle = The process handle.
 * status = The process exit status (return).
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_processes_wait (unsigned int  handle,
                            int          *status);

/*
 * Terminates a process, waits for its exit and
 * removes the entry in the kernel process table.
 *
 * handle = The process handle.
 * status = The process exit status (return).
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_processes_terminate (unsigned int  handle,
                                 int          *status);

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* Plug-in loader directory. */
extern const char *harness_execdir;

#endif /* HARNESS_PROCESSES_H */

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
