/******************************************************************************
 *
 * Source file for rmixregistryd options (option parser) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* Command line version output. */
#ifdef DEBUG
#define RMIXREGISTRYD_OPTIONS_VERSION RMIX_STRING(\
rmixregistryd.RMIXREGISTRYD_VERSION.debug (RMIX registry daemon in debug mode)\
 RMIXREGISTRYD_PACKAGE_NAME-RMIXREGISTRYD_PACKAGE_VERSION.\
RMIXREGISTRYD_PACKAGE_RELEASE\n\
Copyright (c) 2002-2006 Oak Ridge National Laboratory.\n\
See COPYING in the source distribution top-level directory or package\n\
data directory (RMIXREGISTRYD_DATADIR) for license information.\n\
Report bugs to <RMIXREGISTRYD_PACKAGE_BUGREPORT>.\n)
#else /* DEBUG */
#define RMIXREGISTRYD_OPTIONS_VERSION RMIX_STRING(\
rmixregistryd.RMIXREGISTRYD_VERSION (RMIX registry daemon)\
 RMIXREGISTRYD_PACKAGE_NAME-RMIXREGISTRYD_PACKAGE_VERSION.\
RMIXREGISTRYD_PACKAGE_RELEASE\n\
Copyright (c) 2002-2006 Oak Ridge National Laboratory.\n\
See COPYING in the source distribution top-level directory or package\n\
data directory (RMIXREGISTRYD_DATADIR) for license information.\n\
Report bugs to <RMIXREGISTRYD_PACKAGE_BUGREPORT>.\n)
#endif /* DEBUG */

/* Command line help output. */
#define RMIXREGISTRYD_OPTIONS_HELP "\
Usage:\n\
  rmixregistryd[.debug] [option[ |=][VALUE]] ..\n\
Environment:\n\
  RMIXREGISTRYD_OPTIONS=\"[option[ |=][VALUE]] ..\"\n\
%s\
Configuration file format:\n\
  # Comment until end of line.\n\
  [option[ |=][VALUE]] .. \n\
  [option[ |=][VALUE]] .. \n\
  .. \n\
General options:\n\
  -h,  --help     Print out command line help. No option VALUE.\n\
  -v,  --version  Print out version information. No option VALUE.\n\
Export options:\n\
  -e,  --export   Export the registry using the with VALUE specified\n\
                  parameters. VALUE can be ommitted, where PROTOCOL=RPCX\n\
                  is default. VALUE has the form NAME=VALUE .. . Use a\n\
                  leading \\ for any \\, = and space that is part of NAME\n\
                  or VALUE. Enclose multiple parameters in quotes, e.g.\n\
                  'NAME1=VALUE1 NAME2=VALUE2'.\n\
                  Multiple -e options result in multiple exports. However,\n\
                  parameters must be different. Ommiting the -e option\n\
                  defaults to -e PROTOCOL=RPCX.\n\
Program options:\n\
  -D,  --daemon   Set the flag to run as daemon. 0 <= VALUE <= 1,\n\
                  where 1 is default. Omitting VALUE defaults to 1.\n\
  -d,  --debug    Set the flag to run in debug mode. 0 <= VALUE <= 1,\n\
                  where 1 is default in debug mode and 0 is default in.\n\
                  non-debug mode. Omitting VALUE defaults to 1.\n\
Thread pool options:\n\
  -m,  --minimum  Set the minimum thread count of the thread pool to\n\
                  VALUE. VALUE cannot be ommitted, where 5 is default.\n\
                  VALUE must be less than current maximum thread count.\n\
  -M,  --maximum  Set the maximum thread count of the thread pool to\n\
                  VALUE. VALUE cannot be ommitted, where 25 is default.\n\
                  VALUE must be greater than current minimum thread count.\n\
  -T,  --timeout  Set the exiting thread timeout of the thread pool to\n\
                  VALUE. VALUE cannot be ommitted, where 2 is default.\n\
  -q,  --queue    Set the maximum job count of the thread pool to VALUE.\n\
                  VALUE cannot be 0 or ommitted, where 1024 is default.\n\
  -w,  --wait     Set the waiting thread timeout of the thread pool to\n\
                  VALUE. VALUE cannot be ommitted, where 120 is default.\n\
"

#define RMIXREGISTRYD_OPTIONS_HELP_FILES RMIX_STRING(\
Directories:\n\
\40\40Package:\40 RMIXREGISTRYD_DATADIR\n\
Configuration files:\n\
\40\40System:\40\40 RMIXREGISTRYD_SYSCONFDIR/RMIXREGISTRYD_CONFFILE\n\
\40\40User:\40\40\40\40 \
$HOME/RMIXREGISTRYD_USRCONFDIR/RMIXREGISTRYD_CONFFILE\n\
)

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include "rmixregistryd.h"

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* Option parser data type. */
typedef struct {
  unsigned int                    daemon;        /* daemon flag            */
  unsigned int                    debug;         /* debug flag             */
  struct {
    unsigned int                  minimum;       /* minimum thread count   */
    unsigned int                  maximum;       /* maximum thread count   */
    unsigned int                  timeout;       /* exiting thread timeout */
  }                               pool;          /* thread pool            */
  struct {
    unsigned int                  maximum;       /* maximum job count      */
    unsigned int                  timeout;       /* waiting thread timeout */
  }                               jobs;          /* job queue              */
  rmix_parameters_t             **parameters;    /* export parameters      */
} rmixregistryd_options_data_t;

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Parses a configuration file.
 *
 * path   = The configuration file path.
 * strict = The strict flag indicating if the file must exist.
 * argv   = The argument vector (argument[0] is the program name).
 * return = 0 on success or -1 on any error with errno set appropriately.
 *          May exit with EXIT_SUCCESS if an option requires it and may exit
 *          with EXIT_FAILURE for unrecognized options and missing values.
 */
int rmixregistryd_options_file_parse (const char    *path,
                                      unsigned int   strict,
                                      const char   **argv);

/*
 * Parses the environment options.
 *
 * envp   = The variable vector.
 * argv   = The argument vector (argument[0] is the program name).
 * return = 0 on success or -1 on any error with errno set appropriately.
 *          May exit with EXIT_SUCCESS if an option requires it and may exit
 *          with EXIT_FAILURE for unrecognized options and missing values.
 */
int rmixregistryd_options_envp_parse (const char **envp,
                                      const char **argv);

/*
 * Parses the command-line options.
 *
 * argc   = The argument vector count.
 * argv   = The argument vector (argument[0] is the program name).
 * envp   = The variable vector.
 * return = 0 on success or -1 on any error with errno set appropriately.
 *          May exit with EXIT_SUCCESS if an option requires it and may exit
 *          with EXIT_FAILURE for unrecognized options and missing values.
 */
int rmixregistryd_options_argv_parse (int          argc,
                                      const char **argv);

/*
 * Applies internal options, such as the debug flag.
 *
 * argv   = The argument vector (argument[0] is the program name).
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmixregistryd_options_apply (const char **argv);

/*
 * Prints out a message to stdout and exits with an exit status.
 *
 * argv    = The argument vector (argument[0] is the program name).
 * format  = The message format.
 * message = The message.
 * status  = The exit status.
 */
void rmixregistryd_options_exit (const char **argv,
                                 const char  *format,
                                 const char  *message,
                                 int          status);

/*
 * Adds an export parameter list to the export parameters option.
 *
 * pairs  = The export parameter list name-value pairs string. Each pair
 *          must be in the form <name>=<value>. Use a leading '\' for '='
 *          and spaces that are part of name or value and for other '\'s.
 *          For example, 'name1\=\ =\=value1\\ name2\ =value2' means
 *          'name1= ' and '=value1\', and 'name 2' and 'value2'. The name-
 *          value pairs string may contain leading and trailing spaces, and
 *          multiple consecutive separating spaces.
 * argv   = The argument vector (argument[0] is the program name).
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmixregistryd_options_parameters_add (const char  *pairs,
                                          const char **argv);

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* Option parser data. */
rmixregistryd_options_data_t rmixregistryd_options_data = {
  1,                           /* daemon flag              */
#ifdef DEBUG
  1,                           /* debug flag               */
#else /* DEBUG */
  0,                           /* debug flag               */
#endif /* DEBUG */
  {
    RMIX_THREADS_POOL_MINIMUM, /* minimum thread count     */
    RMIX_THREADS_POOL_MAXIMUM, /* maximum thread count     */
    RMIX_THREADS_POOL_TIMEOUT  /* exiting thread timeout   */
  },                           /* thread pool              */
  {
    RMIX_THREADS_JOBS_MAXIMUM, /* maximum job count        */
    RMIX_THREADS_JOBS_TIMEOUT  /* waiting thread timeout   */
  },                           /* job queue                */
  NULL                         /* export parameters        */
};

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Initializes the option parser.
 *
 * argc   = The argument vector count.
 * argv   = The argument vector (argument[0] is the program name).
 * envp   = The variable vector.
 * return = 0 on success or -1 on any error with errno set appropriately.
 *          May exit with EXIT_SUCCESS if an option requires it and may exit
 *          with EXIT_FAILURE for unrecognized options and missing values.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryd_options_init"
int rmixregistryd_options_init (const int    argc,
                                const char **argv,
                                const char **envp) {
  char         *home, *file, *path;
  unsigned int  homelen, filelen;
  RMIX_LOG((RMIX_INFO(option parser startup)))
#ifdef DEBUG
  /* Check argv parameter. */
  if (NULL == argv) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argv parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check argv[0] parameter. */
  if (NULL == argv[0]) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argv[0] parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check envp parameter. */
  if (NULL == envp) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(envp parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Parse system configuration file. */
  if (0 != rmixregistryd_options_file_parse(RMIX_STRING(\
             RMIXREGISTRYD_SYSCONFDIR/RMIXREGISTRYD_CONFFILE), 0, argv)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to parse system configuration file)))
    /* Finalize option parser. */
    if (0 != rmixregistryd_options_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize option parser)))
    }
    errno = errno2;
    return -1;
  }
  if (NULL != (home = getenv("HOME"))) {
    file    = RMIX_STRING(/RMIXREGISTRYD_USRCONFDIR/RMIXREGISTRYD_CONFFILE);
    homelen = strlen(home);
    filelen = strlen(file) + 1;
    /* Allocate user configuration file path. */
    if (NULL == (path = malloc((homelen + filelen) * sizeof(char)))) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to allocate user configuration file path)))
      /* Finalize option parser. */
      if (0 != rmixregistryd_options_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize option parser)))
      }
      errno = errno2;
      return -1;
    }
    /* Compose user configuration file path. */
    memcpy(path          , home, homelen);
    memcpy(path + homelen, file, filelen);
    /* Parse user configuration file. */
    if (0 != rmixregistryd_options_file_parse(path, 0, argv)) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to parse user configuration file)))
      /* Deallocate user configuration file path. */
      free(path);
      /* Finalize option parser. */
      if (0 != rmixregistryd_options_fini()) {
        RMIX_LOG((RMIX_WARN(unable to finalize option parser)))
      }
      errno = errno2;
      return -1;
    }
    /* Deallocate user configuration file path. */
    free(path);
  }
  /* Parse environment options. */
  if (0 != rmixregistryd_options_envp_parse(envp, argv)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to parse environment options)))
    /* Finalize option parser. */
    if (0 != rmixregistryd_options_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize option parser)))
    }
    errno = errno2;
    return -1;
  }
  /* Parse command-line options. */
  if (0 != rmixregistryd_options_argv_parse(argc, argv)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to parse command-line options)))
    /* Finalize option parser. */
    if (0 != rmixregistryd_options_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize option parser)))
    }
    errno = errno2;
    return -1;
  }
  /* Apply internal options. */
  if (0 != rmixregistryd_options_apply(argv)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to apply non-runtime options)))
    /* Finalize option parser. */
    if (0 != rmixregistryd_options_fini()) {
      RMIX_LOG((RMIX_WARN(unable to finalize option parser)))
    }
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Finalizes the option parser.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryd_options_fini"
int rmixregistryd_options_fini () {
  int          result, errno2;
  unsigned int index;
  RMIX_LOG((RMIX_INFO(option parser shutdown)))
  result = 0;
  errno2 = 0;
  /* Destroy export parameters parameter lists. */
  for (index = 0; ((NULL != rmixregistryd_options_data.parameters       )&&
                   (NULL != rmixregistryd_options_data.parameters[index]));
       index++) {
    /* Destroy parameter list. */
    if (0 != rmix_parameters_destroy(
               &rmixregistryd_options_data.parameters[index])) {
      errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to destroy parameter list)))
      result = -1;
    }
  }
  /* Deallocate export parameters. */
  if (NULL != rmixregistryd_options_data.parameters) {
    free(rmixregistryd_options_data.parameters);
    rmixregistryd_options_data.parameters = NULL;
  }
  if (0 != errno2) {
    errno = errno2;
  }
  return result;
}

/*
 * Returns the daemon flag.
 *
 * daemon = The daemon flag (return).
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryd_options_daemon_get"
int rmixregistryd_options_daemon_get (unsigned int *daemon) {
#ifdef DEBUG
  /* Check daemon parameter. */
  if (NULL == daemon) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(daemon parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Save daemon flag. */
  *daemon = rmixregistryd_options_data.daemon;
  return 0;
}

/*
 * Gets the minimum thread count of the thread pool.
 *
 * minimum = The minimum thread count (return).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryd_options_pool_minimum_get"
int rmixregistryd_options_pool_minimum_get (unsigned int *minimum) {
#ifdef DEBUG
  /* Check minimum parameter. */
  if (NULL == minimum) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(minimum parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Save minimum thread count. */
  *minimum = rmixregistryd_options_data.pool.minimum;
  return 0;
}

/*
 * Gets the maximum thread count of the thread pool.
 *
 * maximum = The maximum thread count (return).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryd_options_pool_maximum_get"
int rmixregistryd_options_pool_maximum_get (unsigned int *maximum) {
#ifdef DEBUG
  /* Check maximum parameter. */
  if (NULL == maximum) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(maximum parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Save maximum thread count. */
  *maximum = rmixregistryd_options_data.pool.maximum;
  return 0;
}

/*
 * Gets the exiting thread timeout of the thread pool.
 *
 * timeout = The exiting thread timeout (return).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryd_options_pool_timeout_get"
int rmixregistryd_options_pool_timeout_get (unsigned int *timeout) {
#ifdef DEBUG
  /* Check timeout parameter. */
  if (NULL == timeout) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(timeout parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Save exiting thread timeout. */
  *timeout = rmixregistryd_options_data.pool.timeout;
  return 0;
}

/*
 * Gets the maximum job count of the thread pools job queue.
 *
 * maximum = The maximum job count (return).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryd_options_jobs_maximum_get"
int rmixregistryd_options_jobs_maximum_get (unsigned int *maximum) {
#ifdef DEBUG
  /* Check maximum parameter. */
  if (NULL == maximum) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(maximum parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Save maximum job count. */
  *maximum = rmixregistryd_options_data.jobs.maximum;
  return 0;
}

/*
 * Gets the waiting thread timeout of the thread pools job queue.
 *
 * timeout = The waiting thread timeout (return).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryd_options_jobs_timeout_get"
int rmixregistryd_options_jobs_timeout_get (unsigned int *timeout) {
#ifdef DEBUG
  /* Check timeout parameter. */
  if (NULL == timeout) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(timeout parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Save waiting thread timeout. */
  *timeout = rmixregistryd_options_data.jobs.timeout;
  return 0;
}

/*
 * Gets the export parameters.
 *
 * parameters = The export parameters (return).
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryd_options_parameters_get"
int rmixregistryd_options_parameters_get
      (const rmix_parameters_t ***parameters) {
#ifdef DEBUG
  /* Check parameters parameter. */
  if (NULL == parameters) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(parameters parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Save export parameters. */
  *parameters = (const rmix_parameters_t**)
                  rmixregistryd_options_data.parameters;
  return 0;
}

/*
 * Parses a configuration file.
 *
 * path   = The configuration file path.
 * strict = The strict flag indicating if the file must exist.
 * argv   = The argument vector (argument[0] is the program name).
 * return = 0 on success or -1 on any error with errno set appropriately.
 *          May exit with EXIT_SUCCESS if an option requires it and may exit
 *          with EXIT_FAILURE for unrecognized options and missing values.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryd_options_file_parse"
int rmixregistryd_options_file_parse (const char    *path,
                                      unsigned int   strict,
                                      const char   **argv) {
  int           file, bytes;
  char         *content, *options, *token, **argv1, **argv2;
  off_t         offset;
  unsigned int  length, index, mark, argc1, argc2;
#ifdef DEBUG
  /* Check path parameter. */
  if (NULL == path) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(path parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check argv parameter. */
  if (NULL == argv) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argv parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check argv[0] parameter. */
  if (NULL == argv[0]) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argv[0] parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Open configuration file. */
  if (-1 == (file = open(path, O_RDONLY))) {
    if (1 == strict) {
      /* Exit with error message. */
      rmixregistryd_options_exit(argv,
        "ERROR: Unable to open configuration file \'%s\'.\n",
        path, EXIT_FAILURE);
    } else {
      RMIX_LOG((RMIX_INFO(unable to open configuration file '%s'), path))
      return 0;
    }
  }
  /* Seek to end of configuration file. */
  if (-1 == (offset = lseek(file, 0, SEEK_END))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to seek to end of configuration file)))
    /* Close configuration file. */
    if (0 != close(file)) {
      RMIX_LOG((RMIX_WARN(unable to close configuration file)))
    }
    errno = errno2;
    return -1;
  }
  /* Calculate configuration file length. */
  length = offset;
  /* Seek to start of configuration file. */
  if (-1 == lseek(file, 0, SEEK_SET)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to seek to start of configuration file)))
    /* Close configuration file. */
    if (0 != close(file)) {
      RMIX_LOG((RMIX_WARN(unable to close configuration file)))
    }
    errno = errno2;
    return -1;
  }
  /* Check for empty configuration file. */
  if (0 == length) {
    /* Close configuration file. */
    if (0 != close(file)) {
      RMIX_LOG((RMIX_WARN(unable to close configuration file)))
    }
    return 0;
  }
  /* Allocate configuration file content. */
  if (NULL == (content = malloc(length + 1))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate configuration file content)))
    /* Close configuration file. */
    if (0 != close(file)) {
      RMIX_LOG((RMIX_WARN(unable to close configuration file)))
    }
    errno = errno2;
    return -1;
  }
  /* Read configuration file content. */
  for (index = 0; index < length; index += bytes) {
    /* Read some data. */
    if (-1 == (bytes = read(file, content + index, length - index))) {
      /* Check for other than interruption. */
      if ((errno != EINTR)&&(errno != EAGAIN)) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to read from configuration file)))
        /* Close configuration file. */
        if (0 != close(file)) {
          RMIX_LOG((RMIX_WARN(unable to close configuration file)))
        }
        errno = errno2;
        return -1;
      }
    }
  }
  content[length] = '\0';
  /* Close configuration file. */
  if (0 != close(file)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to close configuration file)))
    /* Deallocate configuration file content. */
    free(content);
    errno = errno2;
    return -1;
  }
  /* Remove any comments from configuration file content. */
  for (index = 0; index < length; index++) {
    if ('#' == content[index]) {
      mark = index;
      for (index++; index < length && '\n' != content[index]; index++);
      memmove(content + mark,
              content + index, (length - index + 1) * sizeof(char));
      length -= (index - mark);
      index   = mark;
    }
  }
  /* Replace any new lines with spaces in configuration file content. */
  for (index = 0; index < length; index++) {
    if ('\n' == content[index]) {
      content[index] = ' ';
    }
  }
  argc1 = 1;
  /* Allocate argument vector. */
  if (NULL == (argv1 = (char**)malloc(sizeof(char*)))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate argument vector)))
    /* Deallocate configuration file content. */
    free(content);
    errno = errno2;
    return -1;
  }
  /* Duplicate argument 0. */
  if (NULL == (argv1[0] = strdup(argv[0]))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate argument 0)))
    /* Deallocate argument vector. */
    free(argv1);
    /* Deallocate configuration file content. */
    free(content);
    errno = errno2;
    return -1;
  }
  /* Remove all leading spaces. */
  for (options = content; '\0' != options[0] && ' ' == options[0]; options++);
  while ('\0' != options[0]) {
    /* Check for ' quoting. */
    if ('\'' == options[0]) {
      /* Search for next ' quote that is not protected by \. */
      for (token = options + 1;
           ( '\0' != token[0])&&
           ((' '  != token[1])||('\'' != token[0])||('\\' == token[-1]))&&
           (('\0' != token[1])||('\'' != token[0])||('\\' == token[-1]));
           token++);
      /* Check for missing ' quote. */
      if ('\0' == token[0]) {
        /* Deallocate arguments. */
        while (0 != argc1) {
          argc1--;
          free(argv1[argc1]);
        }
        /* Deallocate argument vector. */
        free(argv1);
        /* Deallocate configuration file content. */
        free(content);
        /* Exit with error message. */
        rmixregistryd_options_exit(argv,
          "ERROR: Missing closing quote in configuration file '%s'.\n",
          path, EXIT_FAILURE);
      }
    /* Check for " quoting. */
    } else if ('\"' == options[0]) {
      /* Search for next " quote that is not protected by \. */
      for (token = options + 1;
           ( '\0' != token[0])&&
           ((' '  != token[1])||('\"' != token[0])||('\\' == token[-1]))&&
           (('\0' != token[1])||('\"' != token[0])||('\\' == token[-1]));
           token++);
      /* Check for missing " quote. */
      if ('\0' == token[0]) {
        /* Deallocate arguments. */
        while (0 != argc1) {
          argc1--;
          free(argv1[argc1]);
        }
        /* Deallocate argument vector. */
        free(argv1);
        /* Deallocate configuration file content. */
        free(content);
        /* Exit with error message. */
        rmixregistryd_options_exit(argv,
          "ERROR: Missing closing quote in configuration file '%s'.\n",
          path, EXIT_FAILURE);
      }
    } else {
      /* Search for next space. */
      for (token = options; '\0' != token[0] && ' ' != token[0]; token++);
    }
    /* Increase argument vector. */
    argc2 = argc1;
    argc1++;
    /* Reallocate argument vector. */
    if (NULL == (argv2 = (char**)realloc(argv1, argc1 * sizeof(char*)))) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to reallocate argument vector)))
      /* Deallocate arguments. */
      while (0 != argc2) {
        argc2--;
          free(argv1[argc2]);
      }
      /* Deallocate argument vector. */
      free(argv1);
      /* Deallocate configuration file content. */
      free(content);
      errno = errno2;
      return -1;
    }
    argv1 = argv2;
    /* Duplicate argument. */
    if (NULL == (argv1[argc2] = strndup(options, token - options))) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to duplicate argument)))
      /* Deallocate arguments. */
      while (0 != argc2) {
        argc2--;
        free(argv1[argc2]);
      }
      /* Deallocate argument vector. */
      free(argv1);
      /* Deallocate configuration file content. */
      free(content);
      errno = errno2;
      return -1;
    }
    options = token;
    if ('\0' != options[0]) {
      options++;
    }
    /* Remove all leading spaces. */
    for (; '\0' != options[0] && ' ' == options[0]; options++);
  }
  /* Parse environment options. */
  if (0 != rmixregistryd_options_argv_parse(argc1, (const char**)argv1)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to parse configuration file options)))
    /* Deallocate arguments. */
    while (0 != argc1) {
      argc1--;
      free(argv1[argc1]);
    }
    /* Deallocate argument vector. */
    free(argv1);
    /* Deallocate configuration file content. */
    free(content);
    errno = errno2;
    return -1;
  }
  /* Deallocate arguments. */
  while (0 != argc1) {
    argc1--;
    free(argv1[argc1]);
  }
  /* Deallocate argument vector. */
  free(argv1);
  /* Deallocate configuration file content. */
  free(content);
  return 0;
}

/*
 * Parses the environment options.
 *
 * envp   = The variable vector.
 * argv   = The argument vector (argument[0] is the program name).
 * return = 0 on success or -1 on any error with errno set appropriately.
 *          May exit with EXIT_SUCCESS if an option requires it and may exit
 *          with EXIT_FAILURE for unrecognized options and missing values.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryd_options_envp_parse"
int rmixregistryd_options_envp_parse (const char **envp,
                                      const char **argv) {
  char         **argv1,   **argv2;
  const char    *options,  *token;
  unsigned int   index, argc1, argc2;
#ifdef DEBUG
  /* Check envp parameter. */
  if (NULL == envp) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(envp parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check argv parameter. */
  if (NULL == argv) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argv parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check argv[0] parameter. */
  if (NULL == argv[0]) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argv[0] parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  for (index = 0; NULL != envp[index]; index++) {
    if (0 == strncmp(envp[index], "RMIXREGISTRYD_OPTIONS=", 17)) {
      argc1   = 1;
      options = envp[index] + 17;
      /* Allocate argument vector. */
      if (NULL == (argv1 = (char**)malloc(sizeof(char*)))) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to allocate argument vector)))
        errno = errno2;
        return -1;
      }
      /* Duplicate argument 0. */
      if (NULL == (argv1[0] = strdup(argv[0]))) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to duplicate argument 0)))
        /* Deallocate argument vector. */
        free(argv1);
        errno = errno2;
        return -1;
      }
      /* Remove all leading spaces. */
      for (; '\0' != options[0] && ' ' == options[0]; options++);
      while ('\0' != options[0]) {
        /* Check for ' quoting. */
        if ('\'' == options[0]) {
          /* Search for next ' quote that is not protected by \. */
          for (token = options + 1;
               ( '\0' != token[0])&&
               ((' '  != token[1])||('\'' != token[0])||('\\' == token[-1]))&&
               (('\0' != token[1])||('\'' != token[0])||('\\' == token[-1]));
               token++);
          /* Check for missing ' quote. */
          if ('\0' == token[0]) {
            /* Deallocate arguments. */
            while (0 != argc1) {
              argc1--;
              free(argv1[argc1]);
            }
            /* Deallocate argument vector. */
            free(argv1);
            /* Exit with error message. */
            rmixregistryd_options_exit(argv,
              "ERROR: Missing closing quote in %s variable.\n",
              "RMIXREGISTRYD_OPTIONS", EXIT_FAILURE);
          }
        /* Check for " quoting. */
        } else if ('\"' == options[0]) {
          /* Search for next " quote that is not protected by \. */
          for (token = options + 1;
               ( '\0' != token[0])&&
               ((' '  != token[1])||('\"' != token[0])||('\\' == token[-1]))&&
               (('\0' != token[1])||('\"' != token[0])||('\\' == token[-1]));
               token++);
          /* Check for missing " quote. */
          if ('\0' == token[0]) {
            /* Deallocate arguments. */
            while (0 != argc1) {
              argc1--;
              free(argv1[argc1]);
            }
            /* Deallocate argument vector. */
            if (NULL != argv1) {
              free(argv1);
            }
            /* Exit with error message. */
            rmixregistryd_options_exit(argv,
              "ERROR: Missing closing quote in %s variable.\n",
              "RMIXREGISTRYD_OPTIONS", EXIT_FAILURE);
          }
        } else {
          /* Search for next space. */
          for (token = options; '\0' != token[0] && ' ' != token[0]; token++);
        }
        /* Increase argument vector. */
        argc2 = argc1;
        argc1++;
        /* Reallocate argument vector. */
        if (NULL == (argv2 = (char**)realloc(argv1, argc1 * sizeof(char*)))) {
          int errno2 = errno;
          RMIX_LOG((RMIX_WARN(unable to reallocate argument vector)))
          /* Deallocate arguments. */
          while (0 != argc2) {
            argc2--;
            free(argv1[argc2]);
          }
          /* Deallocate argument vector. */
          free(argv1);
          errno = errno2;
          return -1;
        }
        argv1 = argv2;
        /* Duplicate argument. */
        if (NULL == (argv1[argc2] = strndup(options, token - options))) {
          int errno2 = errno;
          RMIX_LOG((RMIX_WARN(unable to duplicate argument)))
          /* Deallocate arguments. */
          while (0 != argc2) {
            argc2--;
            free(argv1[argc2]);
          }
          /* Deallocate argument vector. */
          free(argv1);
          errno = errno2;
          return -1;
        }
        options = token;
        if ('\0' != options[0]) {
          options++;
        }
        /* Remove all leading spaces. */
        for (; '\0' != options[0] && ' ' == options[0]; options++);
      }
      /* Parse environment options. */
      if (0 != rmixregistryd_options_argv_parse(argc1, (const char**)argv1)) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to parse environment options)))
        /* Deallocate arguments. */
        while (0 != argc1) {
          argc1--;
          free(argv1[argc1]);
        }
        /* Deallocate argument vector. */
        free(argv1);
        errno = errno2;
        return -1;
      }
      /* Deallocate arguments. */
      while (0 != argc1) {
        argc1--;
        if (NULL != argv1[argc1]) {
          free(argv1[argc1]);
        }
      }
      /* Deallocate argument vector. */
      free(argv1);
    }
  }
  return 0;
}

/*
 * Parses the command-line options.
 *
 * argc   = The argument vector count.
 * argv   = The argument vector (argument[0] is the program name).
 * envp   = The variable vector.
 * return = 0 on success or -1 on any error with errno set appropriately.
 *          May exit with EXIT_SUCCESS if an option requires it and may exit
 *          with EXIT_FAILURE for unrecognized options and missing values.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryd_options_argv_parse"
int rmixregistryd_options_argv_parse (int          argc,
                                      const char **argv) {
  unsigned int index;
#ifdef DEBUG
  /* Check argv parameter. */
  if (NULL == argv) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argv parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check argv[0] parameter. */
  if (NULL == argv[0]) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argv[0] parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  for (index = 1; index < argc; index++) {
    if ((0 == strcmp (argv[index],  "-D"     ))||
        (0 == strcmp (argv[index], "--daemon"))) {
      /* Check for next argument. */
      if ((index + 1 < argc)&&('-' != argv[index + 1][0])) {
        index++;
        if (0 == strcmp(argv[index], "0")) {
          rmixregistryd_options_data.daemon = 0;
        } else if (0 == strcmp(argv[index], "1")) {
          rmixregistryd_options_data.daemon = 1;
        } else {
          /* Exit with error message. */
          rmixregistryd_options_exit(argv,
            "ERROR: Illegal value for '%s' option .\n",
            argv[index - 1], EXIT_FAILURE);
        }
      } else {
        rmixregistryd_options_data.daemon = 1;
      }
    } else if ((0 == strcmp (argv[index],  "-D=0"     ))||
               (0 == strcmp (argv[index], "--daemon=0"))) {
      rmixregistryd_options_data.daemon = 0;
    } else if ((0 == strcmp (argv[index],  "-D=1"     ))||
               (0 == strcmp (argv[index], "--daemon=1"))) {
      rmixregistryd_options_data.daemon = 1;
    } else if ((0 == strcmp (argv[index],  "-D="     ))||
               (0 == strcmp (argv[index], "--daemon="))) {
      rmixregistryd_options_data.daemon = 1;
    } else if ((0 == strncmp(argv[index],  "-D="     , 3))||
               (0 == strncmp(argv[index], "--daemon=", 9))) {
      /* Exit with error message. */
      rmixregistryd_options_exit(argv,
        "ERROR: Illegal value for '%s' option .\n", argv[index], EXIT_FAILURE);
    } else if ((0 == strcmp (argv[index],  "-d"    ))||
               (0 == strcmp (argv[index], "--debug"))) {
      /* Check for next argument. */
      if ((index + 1 < argc)&&('-' != argv[index + 1][0])) {
        index++;
        if (0 == strcmp(argv[index], "0")) {
          rmixregistryd_options_data.debug = 0;
        } else if (0 == strcmp(argv[index], "1")) {
          rmixregistryd_options_data.debug = 1;
        } else {
          /* Exit with error message. */
          rmixregistryd_options_exit(argv,
            "ERROR: Illegal value for '%s' option .\n",
            argv[index - 1], EXIT_FAILURE);
        }
      } else {
        rmixregistryd_options_data.debug = 1;
      }
    } else if ((0 == strcmp (argv[index],  "-d=0"    ))||
               (0 == strcmp (argv[index], "--debug=0"))) {
      rmixregistryd_options_data.debug = 0;
    } else if ((0 == strcmp (argv[index],  "-d=1"    ))||
               (0 == strcmp (argv[index], "--debug=1"))) {
      rmixregistryd_options_data.debug = 1;
    } else if ((0 == strcmp (argv[index],  "-d="    ))||
               (0 == strcmp (argv[index], "--debug="))) {
      rmixregistryd_options_data.debug = 1;
    } else if ((0 == strncmp(argv[index],  "-d="    , 3))||
               (0 == strncmp(argv[index], "--debug=", 8))) {
      /* Exit with error message. */
      rmixregistryd_options_exit(argv,
        "ERROR: Illegal value for '%s' option .\n", argv[index], EXIT_FAILURE);
    } else if ((0 == strcmp (argv[index],  "-e"     ))||
               (0 == strcmp (argv[index], "--export"))) {
      /* Check for next argument. */
      if ((index + 1 < argc)&&('-' != argv[index + 1][0])) {
        index++;
        if (0 != rmixregistryd_options_parameters_add(argv[index], argv)) {
          int errno2 = errno;
          RMIX_LOG((RMIX_WARN(unable to add export parameters)))
          errno = errno2;
          return -1;
        }
      } else {
        if (0 != rmixregistryd_options_parameters_add("", argv)) {
          int errno2 = errno;
          RMIX_LOG((RMIX_WARN(unable to add export parameters)))
          errno = errno2;
          return -1;
        }
      }
    } else if ( 0 == strncmp(argv[index], "-e=", 3)) {
      if (0 != rmixregistryd_options_parameters_add(argv[index] + 3, argv)) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to add export parameters)))
        errno = errno2;
        return -1;
      }
    } else if ( 0 == strncmp(argv[index], "--export=", 9)) {
      if (0 != rmixregistryd_options_parameters_add(argv[index] + 9, argv)) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to add export parameters)))
        errno = errno2;
        return -1;
      }
    } else if ((0 == strcmp (argv[index],  "-f"   ))||
               (0 == strcmp (argv[index], "--file"))) {
      /* Check for next argument. */
      index++;
      if (index < argc) {
        /* Parse configuration file. */
        if (0 != rmixregistryd_options_file_parse(argv[index], 1, argv)) {
          int errno2 = errno;
          RMIX_LOG((RMIX_WARN(unable to parse configuration file)))
          errno = errno2;
          return -1;
        }
      } else {
        /* Exit with error message. */
        rmixregistryd_options_exit(argv,
          "ERROR: Missing value for '%s' option.\n",
          argv[index - 1], EXIT_FAILURE);
      }
    } else if ( 0 == strncmp(argv[index],  "-f="   , 3)) {
      /* Check for value in argument. */
      if (3 == strlen(argv[index])) {
        /* Exit with error message. */
        rmixregistryd_options_exit(argv,
          "ERROR: Missing value in '%s' option.\n",
          argv[index], EXIT_FAILURE);
      }
      /* Parse configuration file. */
      if (0 != rmixregistryd_options_file_parse(argv[index] + 3, 1, argv)) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to parse configuration file)))
        errno = errno2;
        return -1;
      }
    } else if ( 0 == strncmp(argv[index], "--file=", 7)) {
      /* Check for value in argument. */
      if (7 == strlen(argv[index])) {
        /* Exit with error message. */
        rmixregistryd_options_exit(argv,
          "ERROR: Missing value in '%s' option.\n",
          argv[index], EXIT_FAILURE);
      }
      /* Parse configuration file. */
      if (0 != rmixregistryd_options_file_parse(argv[index] + 7, 1, argv)) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to parse configuration file)))
        errno = errno2;
        return -1;
      }
    } else if ((0 == strcmp (argv[index],  "-h"   ))||
               (0 == strcmp (argv[index], "--help"))) {
      /* Exit with help message. */
      rmixregistryd_options_exit(argv, RMIXREGISTRYD_OPTIONS_HELP,
        RMIXREGISTRYD_OPTIONS_HELP_FILES, EXIT_SUCCESS);
    } else if  ((0 == strcmp (argv[index],  "-m"      ))||
                (0 == strcmp (argv[index], "--minimum"))) {
      /* Check for next argument. */
      index++;
      if (index < argc) {
        rmixregistryd_options_data.pool.minimum = atoi(argv[index]);
        /* Check range. */
        if (rmixregistryd_options_data.pool.minimum >
            rmixregistryd_options_data.pool.maximum) {
          /* Exit with error message. */
          rmixregistryd_options_exit(argv,
            "ERROR: Value for '%s' is too big.\n",
            argv[index - 1], EXIT_FAILURE);
        }
      } else {
        /* Exit with error message. */
        rmixregistryd_options_exit(argv,
          "ERROR: Missing value for '%s' option.\n",
          argv[index - 1], EXIT_FAILURE);
      }
    } else if ( 0 == strncmp(argv[index], "-m=", 3)) {
      /* Check for value in argument. */
      if (3 == strlen(argv[index])) {
        /* Exit with error message. */
        rmixregistryd_options_exit(argv,
          "ERROR: Missing value in '%s' option.\n", argv[index], EXIT_FAILURE);
      }
      rmixregistryd_options_data.pool.minimum = atoi(argv[index] + 3);
      /* Check range. */
      if (rmixregistryd_options_data.pool.minimum >
          rmixregistryd_options_data.pool.maximum) {
        /* Exit with error message. */
        rmixregistryd_options_exit(argv, "ERROR: Value for '%s' is too big.\n",
                              argv[index], EXIT_FAILURE);
      }
    } else if ( 0 == strncmp(argv[index], "--minimum=", 10)) {
      /* Check for value in argument. */
      if (10 == strlen(argv[index])) {
        /* Exit with error message. */
        rmixregistryd_options_exit(argv,
          "ERROR: Missing value in '%s' option.\n", argv[index], EXIT_FAILURE);
      }
      rmixregistryd_options_data.pool.minimum = atoi(argv[index] + 10);
      /* Check range. */
      if (rmixregistryd_options_data.pool.minimum >
          rmixregistryd_options_data.pool.maximum) {
        /* Exit with error message. */
        rmixregistryd_options_exit(argv,
          "ERROR: Value for '%s' is too big.\n", argv[index], EXIT_FAILURE);
      }
    } else if  ((0 == strcmp (argv[index],  "-M"      ))||
                (0 == strcmp (argv[index], "--maximum"))) {
      /* Check for next argument. */
      index++;
      if (index < argc) {
        rmixregistryd_options_data.pool.maximum = atoi(argv[index]);
        /* Check range. */
        if (0 == rmixregistryd_options_data.pool.maximum) {
          /* Exit with error message. */
          rmixregistryd_options_exit(argv,
                                "ERROR: Value for '%s' option cannot be 0.\n",
                                argv[index - 1], EXIT_FAILURE);
        }
        if (rmixregistryd_options_data.pool.minimum >
            rmixregistryd_options_data.pool.maximum) {
          /* Exit with error message. */
          rmixregistryd_options_exit(argv,
            "ERROR: Value for '%s' is too small.\n",
            argv[index - 1], EXIT_FAILURE);
        }
      } else {
        /* Exit with error message. */
        rmixregistryd_options_exit(argv,
          "ERROR: Missing value for '%s' option.\n",
          argv[index - 1], EXIT_FAILURE);
      }
    } else if ( 0 == strncmp(argv[index], "-M=", 3)) {
      /* Check for value in argument. */
      if (3 == strlen(argv[index])) {
        /* Exit with error message. */
        rmixregistryd_options_exit(argv, 
          "ERROR: Missing value in '%s' option.\n", argv[index], EXIT_FAILURE);
      }
      rmixregistryd_options_data.pool.maximum = atoi(argv[index] + 3);
      /* Check range. */
      if (0 == rmixregistryd_options_data.pool.maximum) {
        /* Exit with error message. */
        rmixregistryd_options_exit(argv,
          "ERROR: Value for '%s' option cannot be 0.\n",
          argv[index], EXIT_FAILURE);
      }
      if (rmixregistryd_options_data.pool.minimum >
          rmixregistryd_options_data.pool.maximum) {
        /* Exit with error message. */
        rmixregistryd_options_exit(argv,
          "ERROR: Value for '%s' is too small.\n", argv[index], EXIT_FAILURE);
      }
    } else if ( 0 == strncmp(argv[index], "--maximum=", 10)) {
      /* Check for value in argument. */
      if (10 == strlen(argv[index])) {
        /* Exit with error message. */
        rmixregistryd_options_exit(argv,
          "ERROR: Missing value in '%s' option.\n",
          argv[index], EXIT_FAILURE);
      }
      rmixregistryd_options_data.pool.maximum = atoi(argv[index] + 10);
      /* Check range. */
      if (0 == rmixregistryd_options_data.pool.maximum) {
        /* Exit with error message. */
        rmixregistryd_options_exit(argv,
          "ERROR: Value for '%s' option cannot be 0.\n",
          argv[index], EXIT_FAILURE);
      }
      if (rmixregistryd_options_data.pool.minimum >
          rmixregistryd_options_data.pool.maximum) {
        /* Exit with error message. */
        rmixregistryd_options_exit(argv,
          "ERROR: Value for '%s' is too small.\n", argv[index], EXIT_FAILURE);
      }
    } else if  ((0 == strcmp (argv[index],  "-q"    ))||
                (0 == strcmp (argv[index], "--queue"))) {
      /* Check for next argument. */
      index++;
      if (index < argc) {
        rmixregistryd_options_data.jobs.maximum = atoi(argv[index]);
        /* Check range. */
        if (0 == rmixregistryd_options_data.jobs.maximum) {
          /* Exit with error message. */
          rmixregistryd_options_exit(argv,
            "ERROR: Value for '%s' option cannot be 0.\n",
            argv[index - 1], EXIT_FAILURE);
        }
      } else {
        /* Exit with error message. */
        rmixregistryd_options_exit(argv,
          "ERROR: Missing value for '%s' option.\n",
          argv[index - 1], EXIT_FAILURE);
      }
    } else if ( 0 == strncmp(argv[index], "-q=", 3)) {
      /* Check for value in argument. */
      if (3 == strlen(argv[index])) {
        /* Exit with error message. */
        rmixregistryd_options_exit(argv,
          "ERROR: Missing value in '%s' option.\n", argv[index], EXIT_FAILURE);
      }
      rmixregistryd_options_data.jobs.maximum = atoi(argv[index] + 3);
      /* Check range. */
      if (0 == rmixregistryd_options_data.jobs.maximum) {
        /* Exit with error message. */
        rmixregistryd_options_exit(argv,
          "ERROR: Value for '%s' option cannot be 0.\n",
          argv[index], EXIT_FAILURE);
      }
    } else if ( 0 == strncmp(argv[index], "--queue=", 8)) {
      /* Check for value in argument. */
      if (8 == strlen(argv[index])) {
        /* Exit with error message. */
        rmixregistryd_options_exit(argv,
          "ERROR: Missing value in '%s' option.\n", argv[index], EXIT_FAILURE);
      }
      rmixregistryd_options_data.jobs.maximum = atoi(argv[index] + 8);
      /* Check range. */
      if (0 == rmixregistryd_options_data.jobs.maximum) {
        /* Exit with error message. */
        rmixregistryd_options_exit(argv,
          "ERROR: Value for '%s' option cannot be 0.\n",
          argv[index], EXIT_FAILURE);
      }
    } else if  ((0 == strcmp (argv[index],  "-T"      ))||
                (0 == strcmp (argv[index], "--timeout"))) {
      /* Check for next argument. */
      index++;
      if (index < argc) {
        rmixregistryd_options_data.pool.timeout = atoi(argv[index]);
      } else {
        /* Exit with error message. */
        rmixregistryd_options_exit(argv,
          "ERROR: Missing value for '%s' option.\n",
          argv[index - 1], EXIT_FAILURE);
      }
    } else if ( 0 == strncmp(argv[index], "-T=", 3)) {
      /* Check for value in argument. */
      if (3 == strlen(argv[index])) {
        /* Exit with error message. */
        rmixregistryd_options_exit(argv,
          "ERROR: Missing value in '%s' option.\n", argv[index], EXIT_FAILURE);
      }
      rmixregistryd_options_data.pool.timeout = atoi(argv[index] + 3);
    } else if ( 0 == strncmp(argv[index], "--timeout=", 10)) {
      /* Check for value in argument. */
      if (10 == strlen(argv[index])) {
        /* Exit with error message. */
        rmixregistryd_options_exit(argv,
          "ERROR: Missing value in '%s' option.\n", argv[index], EXIT_FAILURE);
      }
      rmixregistryd_options_data.pool.timeout = atoi(argv[index] + 10);
    } else if ((0 == strcmp (argv[index],  "-v"      ))||
               (0 == strcmp (argv[index], "--version"))) {
      /* Exit with version message. */
      rmixregistryd_options_exit(argv,
        "%s", RMIXREGISTRYD_OPTIONS_VERSION, EXIT_SUCCESS);

    } else if  ((0 == strcmp (argv[index],  "-w"   ))||
                (0 == strcmp (argv[index], "--wait"))) {
      /* Check for next argument. */
      index++;
      if (index < argc) {
        rmixregistryd_options_data.jobs.timeout = atoi(argv[index]);
      } else {
        /* Exit with error message. */
        rmixregistryd_options_exit(argv,
          "ERROR: Missing value for '%s' option.\n",
          argv[index - 1], EXIT_FAILURE);
      }
    } else if ( 0 == strncmp(argv[index], "-w=", 3)) {
      /* Check for value in argument. */
      if (3 == strlen(argv[index])) {
        /* Exit with error message. */
        rmixregistryd_options_exit(argv,
          "ERROR: Missing value in '%s' option.\n", argv[index], EXIT_FAILURE);
      }
      rmixregistryd_options_data.jobs.timeout = atoi(argv[index] + 3);
    } else if ( 0 == strncmp(argv[index], "--wait=", 7)) {
      /* Check for value in argument. */
      if (7 == strlen(argv[index])) {
        /* Exit with error message. */
        rmixregistryd_options_exit(argv,
          "ERROR: Missing value in '%s' option.\n", argv[index], EXIT_FAILURE);
      }
      rmixregistryd_options_data.jobs.timeout = atoi(argv[index] + 7);
    } else {
      /* Exit with error message. */
      rmixregistryd_options_exit(argv, "ERROR: Unknown option '%s'.\n",
                                 argv[index], EXIT_FAILURE);
    }
  }
  return 0;
}

/*
 * Applies internal options, such as the debug flag.
 *
 * argv   = The argument vector (argument[0] is the program name).
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryd_options_apply"
int rmixregistryd_options_apply (const char **argv) {
  char         *argv0;
  unsigned int  length;
#ifdef DEBUG
  /* Check argv parameter. */
  if (NULL == argv) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argv parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check argv[0] parameter. */
  if (NULL == argv[0]) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argv[0] parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
#ifdef DEBUG
  /* Check the debug flag. */
  if (0 == rmixregistryd_options_data.debug) {
    length = strlen(argv[0]) - 6;
    /* Allocate argv[0] duplicate. */
    if (NULL == (argv0 = malloc((length + 1)* sizeof(char)))) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to allocate argv[0] duplicate)))
      errno = errno2;
      return -1;
    }
    memcpy(argv0, argv[0], length);
    argv0[length] = 0;
    /* Execute non-debug version. */
    execv(argv0, (char**)argv);
    RMIX_LOG((RMIX_WARN(unable to execute non-debug version)))
    return -1;
  }
#else /* DEBUG */
  /* Check the debug flag. */
  if (1 == rmixregistryd_options_data.debug) {
    length = strlen(argv[0]);
    /* Allocate argv[0] duplicate. */
    if (NULL == (argv0 = malloc((length  + 7)* sizeof(char)))) {
      return -1;
    }
    memcpy(argv0         , argv[0] , length);
    memcpy(argv0 + length, ".debug", 7);
    execv(argv0, (char**)argv);
    return -1;
  }
#endif /* DEBUG */
  /* Check for empty parameter list. */
  if (NULL == rmixregistryd_options_data.parameters) {
    /* Default to standard export.*/
    if (0 != rmixregistryd_options_parameters_add(
               RMIXREGISTRY_EXPORT_RPC, argv)) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to add export parameters)))
      errno = errno2;
      return -1;
    }
  }
  return 0;
}

/*
 * Prints out a message to stdout and exits with an exit status.
 *
 * argv    = The argument vector (argument[0] is the program name).
 * format  = The message format.
 * message = The message.
 * status  = The exit status.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryd_options_exit"
void rmixregistryd_options_exit (const char **argv,
                                 const char  *format,
                                 const char  *message,
                                 int          status) {
#ifdef DEBUG
  /* Check argv parameter. */
  if (NULL == argv) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argv parameter is null)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    exit(EXIT_FAILURE);
  }
  /* Check argv[0] parameter. */
  if (NULL == argv[0]) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argv[0] parameter is null)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    exit(EXIT_FAILURE);
  }
  /* Check format parameter. */
  if (NULL == format) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(format parameter is null)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    exit(EXIT_FAILURE);
  }
  /* Check message parameter. */
  if (NULL == message) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(message parameter is null)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    exit(EXIT_FAILURE);
  }
#endif /* DEBUG */
  /* Finalize option parser. */
  if (0 != rmixregistryd_options_fini()) {
    RMIX_LOG((RMIX_WARN(unable to finalize option parser)))
    exit(EXIT_FAILURE);
  }
  fprintf(stdout, format, message);
  if (EXIT_FAILURE == status) {
    fprintf(stdout, "Try `%s --help` for more information.\n", argv[0]);
  }
  exit(status);
}

/*
 * Adds an export parameter list to the export parameters option.
 *
 * pairs  = The export parameter list name-value pairs string. Each pair
 *          must be in the form <name>=<value>. Use a leading '\' for '='
 *          and spaces that are part of name or value and for other '\'s.
 *          For example, 'name1\=\ =\=value1\\ name2\ =value2' means
 *          'name1= ' and '=value1\', and 'name 2' and 'value2'. The name-
 *          value pairs string may contain leading and trailing spaces, and
 *          multiple consecutive separating spaces.
 * argv   = The argument vector (argument[0] is the program name).
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmixregistryd_options_parameters_add"
int rmixregistryd_options_parameters_add (const char  *pairs,
                                          const char **argv) {
  int                diff;
  void              *eparameters;
  unsigned int       index;
  rmix_parameters_t *parameters;
#ifdef DEBUG
  /* Check pairs parameter. */
  if (NULL == pairs) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(pairs parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check argv parameter. */
  if (NULL == argv) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(argv parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Check for empty parameter list. */
  if ('\0' == pairs[0]) {
    /* Default to "PROTOCOL=RPCX".*/
    pairs = "PROTOCOL=RPCX";
  }
  /* Create parameter list. */
  if (0 != rmix_parameters_create4(&parameters, pairs)) {
    RMIX_LOG((RMIX_WARN(unable to create parameter list)))
    /* Exit with error message. */
    rmixregistryd_options_exit(argv,
      "ERROR: Unable to create parameter list '%s'.\n", pairs, EXIT_FAILURE);
  }
  /* Compare parameters. */
  for (index = 0; ((NULL != rmixregistryd_options_data.parameters       )&&
                   (NULL != rmixregistryd_options_data.parameters[index]));
       index++) {
    /* Compare parameter lists. */
    if (0 != rmix_parameters_compare(&diff, parameters,
               rmixregistryd_options_data.parameters[index])) {
      RMIX_LOG((RMIX_WARN(unable to compare parameter list)))
      /* Destroy parameter list. */
      if (0 != rmix_parameters_destroy(&parameters)) {
        RMIX_LOG((RMIX_WARN(unable to destroy parameter list)))
      }
      /* Exit with error message. */
      rmixregistryd_options_exit(argv,
        "ERROR: An internal error has occurred.\n", NULL, EXIT_FAILURE);
    }
    /* Check for equal parameter lists. */
    if (0 == diff) {
      /* Destroy parameter list. */
      if (0 != rmix_parameters_destroy(&parameters)) {
        RMIX_LOG((RMIX_WARN(unable to destroy parameter list)))
        /* Exit with error message. */
        rmixregistryd_options_exit(argv,
          "ERROR: An internal error has occurred.\n", NULL, EXIT_FAILURE);
      }
    }
    return 0;
  }
  /* Reallocate export parameters. */
  if (NULL == (eparameters =
               realloc(rmixregistryd_options_data.parameters, (index + 2) *
                sizeof(rmixregistryd_options_data.parameters[0])))) {
    RMIX_LOG((RMIX_WARN(unable to reallocate export parameters)))
    /* Exit with error message. */
    rmixregistryd_options_exit(argv,
      "ERROR: An internal error has occurred.\n", NULL, EXIT_FAILURE);
    return -1;
  }
  rmixregistryd_options_data.parameters = eparameters;
  /* Save parameter list and terminate export parameters. */
  rmixregistryd_options_data.parameters[index    ] = parameters;
  rmixregistryd_options_data.parameters[index + 1] = NULL;
  return 0;
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
